import json
import urllib.request

from vpf_730.fifo_queue import connect
from vpf_730.fifo_queue import Message
from vpf_730.worker import Config
from vpf_730.worker import register


@register
def post_data(msg: Message, cfg: Config) -> None:
    """Send a ``POST`` request based on a :func:`vpf_730.fifo_queue.Message` to
    an http endpoint, which is defined in ``cfg``. The data is send in the body
    as a ``json``. The authorization is done via setting a header:
    ``Authorization: api-key``.

    :param msg: the :func:`vpf_730.fifo_queue.Message` to post (not yet
        serialized)
    :param cfg: the configuration, a ``NamedTuple``
        (:func:`vpf_730.worker.Config`) containing the needed API-key and
        endpoint url
    """
    if cfg.endpoint is None or cfg.api_key is None:
        raise ValueError(
            'no values for endpoint or api_key provided in the cfg object',
        )

    post_data = json.dumps({'data': [msg.blob._asdict()]}).encode()
    req = urllib.request.Request(cfg.endpoint, data=post_data)
    req.add_header('Authorization', cfg.api_key)
    req.add_header('Content-Type', 'application/json')
    urllib.request.urlopen(req)


CREATE_TABLE = '''\
        CREATE TABLE IF NOT EXISTS measurements(
            timestamp INT PRIMARY KEY,
            sensor_id INT NOT NULL,
            last_measurement_period INT,
            time_since_report INT,
            optical_range NUMERIC,
            precipitation_type_msg TEXT,
            obstruction_to_vision TEXT,
            receiver_bg_illumination NUMERIC,
            water_in_precip NUMERIC,
            temp NUMERIC,
            nr_precip_particles INT,
            transmission_eq NUMERIC,
            exco_less_precip_particle NUMERIC,
            backscatter_exco NUMERIC,
            self_test VARCHAR(3),
            total_exco NUMERIC
        )
    '''


@register
def save_locally(msg: Message, cfg: Config) -> None:
    """Save a :func:`vpf_730.fifo_queue.Message` locally to a database that is
    defined in  ``cfg.local_db``.

    :param msg: the :func:`vpf_730.fifo_queue.Message` to save to a database
        (not yet serialized)
    :param cfg: the :func:`vpf_730.worker.Config` containing the local database
        information
    """
    with connect(cfg.local_db) as db:
        db.execute(CREATE_TABLE)
        db.execute(
            '''\
            INSERT INTO measurements(
                timestamp,
                sensor_id,
                last_measurement_period,
                time_since_report,
                optical_range,
                precipitation_type_msg,
                obstruction_to_vision,
                receiver_bg_illumination,
                water_in_precip,
                temp,
                nr_precip_particles,
                transmission_eq,
                exco_less_precip_particle,
                backscatter_exco,
                self_test,
                total_exco
            )
            VALUES (
                :timestamp,
                :sensor_id,
                :last_measurement_period,
                :time_since_report,
                :optical_range,
                :precipitation_type_msg,
                :obstruction_to_vision,
                :receiver_bg_illumination,
                :water_in_precip,
                :temp,
                :nr_precip_particles,
                :transmission_eq,
                :exco_less_precip_particle,
                :backscatter_exco,
                :self_test,
                :total_exco
            )
            ''',
            msg.blob._asdict(),
        )



from __future__ import annotations

import argparse
import os
import time
from argparse import RawDescriptionHelpFormatter
from collections.abc import Sequence
from datetime import datetime
from datetime import timezone
from uuid import uuid4

from vpf_730.fifo_queue import Message
from vpf_730.fifo_queue import Queue
from vpf_730.tasks import post_data
from vpf_730.tasks import save_locally
from vpf_730.vpf_730 import Measurement
from vpf_730.vpf_730 import retry
from vpf_730.vpf_730 import VPF730
from vpf_730.worker import Config
from vpf_730.worker import Worker

# VPF730_SENTRY_DSN env var needs to be set for monitoring
try:
    import sentry_sdk
    sentry_sdk.init(
        dsn=os.environ.get('VPF730_SENTRY_DSN'),
        traces_sample_rate=1.0,
    )
except ImportError:  # pragma: no cover
    pass


@retry(retries=3)
def _get_measurement(vpf730: VPF730) -> Measurement | None:
    return vpf730.measure()


def main_loop(cfg: Config) -> None:
    queue = Queue(cfg.queue_db, keep_msg=1000, prune_interval=100)
    vpf730 = VPF730(cfg.serial_port)
    worker = Worker(queue=queue, cfg=cfg, daemon=True)
    worker.start()
    while True:
        try:
            time.sleep(worker.poll_interval)
            now = datetime.now(timezone.utc)
            if now.second == 0:
                m = _get_measurement(vpf730)
                if m:
                    post = Message(id=uuid4(), task=post_data.__name__, blob=m)
                    local = Message(
                        id=uuid4(),
                        task=save_locally.__name__,
                        blob=m,
                    )
                    queue.put(local)
                    queue.put(post)
                    # we need to sleep a little, to not accidentally create two
                    # messages
                    time.sleep(1)
                else:
                    print('received an empty message')
        except KeyboardInterrupt:
            try:
                print('\nwaiting for worker to finish all tasks...')
                worker.finish_and_join()
                break
            except KeyboardInterrupt:
                worker.running = False
                print('\nworker finishing current task...')
                worker.join()
                break


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        'vpf-730',
        formatter_class=RawDescriptionHelpFormatter,
    )
    parser.epilog = (
        'If no arguments are provided, the configuration will be read from '
        'the environment variables.\n'
        '  - VPF730_LOCAL_DB\n'
        '  - VPF730_QUEUE_DB\n'
        '  - VPF730_PORT\n'
        '  - VPF730_ENDPOINT\n'
        '  - VPF730_API_KEY\n'
        'For variable descriptions see the CLI arguments above'
    )
    cli_config = parser.add_argument_group('config from CLI')
    cli_config.add_argument(
        '--local-db',
        default='~/vpf_730_local.db',
        help='Path to the local database',
    )
    cli_config.add_argument(
        '--queue-db',
        default='~/vpf_730_queue.db',
        help='Path to the queue database',
    )
    cli_config.add_argument(
        '--serial-port',
        help='Serial port the VPF-730 sensor is connected to, e.g /dev/ttyS0',
    )
    cli_config.add_argument(
        '--endpoint',
        help=(
            'API endpoint to send the data to e.g. https://api.example/com/. '
            'The API-Key must be provided as an environment variable '
            'VPF730_API_KEY=mykey'
        ),
    )
    file_config = parser.add_argument_group('config from file')
    file_config.description = (
        'Reads the configuration from a file '
        'and overrides all previous CLI options'
    )
    file_config.add_argument(
        '-c', '--config',
        help='Path to an .ini config file',
    )
    return parser


def main(argv: Sequence[str] | None = None) -> int:
    parser = build_parser()
    args = parser.parse_args(argv)
    if args.serial_port:
        cfg = Config.from_argparse(args=args)
    elif args.config:
        cfg = Config.from_file(args.config)
    else:
        cfg = Config.from_env()

    main_loop(cfg=cfg)
    return 0


if __name__ == '__main__':
    raise SystemExit(main())



from __future__ import annotations

import argparse
import configparser
import os
import textwrap
import threading
import time
import traceback
from collections.abc import Iterable
from collections.abc import Mapping
from typing import Any
from typing import Callable
from typing import NamedTuple

from vpf_730.fifo_queue import Message
from vpf_730.fifo_queue import Queue

try:
    import sentry_sdk
    SENTRY = True
except ImportError:  # pragma: no cover
    SENTRY = False

# this should be generic: https://github.com/python/mypy/issues/11855
"""
Dictionary containing the tasks that are registered. The workers uses the task
name ``my_func.__name__`` to get the corresponding ``Callable``. Tasks may be
added to this using the :func:`register` decorator.
"""
TASKS: dict[str, Callable[[Message, Config], None]] = {}


def register(
        f: Callable[[Message, Config], None],
) -> Callable[[Message, Config], None]:
    """A decorator to register a task for the worker in the :const:`TASK`
    dictionary. The registered function must take a
    :func:`vpf_730.fifo_queue.Message` as the first positional argument and a
    :func:`Config` as the second and must return ``None``. This can be done
    like this:

    .. highlight:: python
    .. code-block:: python

        @register
        def my_func(msg: Message, cfg: Config) -> None:
            ...
            return None

    :param f: a callable taking a :func:`vpf_730.fifo_queue.Message` as the
        first positional argument and a :func:`Config` as the second positional
        argument. The function should return ``None``.

    :return: the function passed to the decorator
    """
    TASKS[f.__name__] = f
    return f


class Config(NamedTuple):
    """A class representing the configuration of the software.

    :param queue_db: path to the sqlite database which is used as a queue
    :param local_db: path to the local sqlite database, where the measurements
        are stored
    :param serial_port: serial port that the VPF-730 sensor is connected to
    :param endpoint: http endpoint where the data should be posted to using
        :func:`vpf_730.tasks.post_data`
    :param api_key: the API-key used to authenticate when sending the ``POST``
        request in :func:`vpf_730.tasks.post_data`
    """
    queue_db: str
    local_db: str
    serial_port: str
    endpoint: str | None = None
    api_key: str | None = None

    @classmethod
    def from_env(cls) -> Config:
        """Constructs a new :func:`Config` from environment variables.

        * ``VPF730_LOCAL_DB`` - path to the sqlite database which is used as  queue
        * ``VPF730_QUEUE_DB`` - path to the local sqlite database, where the measurements are stored
        * ``VPF730_PORT`` - serial port that the VPF-730 sensor is connected to
        * ``VPF730_ENDPOINT`` - http endpoint where the data should be posted to using :func:`vpf_730.tasks.post_data`
        * ``VPF730_API_KEY`` - the API-key used to authenticate when sending the ``POST`` request in :func:`vpf_730.tasks.post_data`

        :return: a new instance of :func:`Config` created from environment
            variables.
        """  # noqa: E501
        return cls(
            local_db=os.environ['VPF730_LOCAL_DB'],
            queue_db=os.environ['VPF730_QUEUE_DB'],
            serial_port=os.environ['VPF730_PORT'],
            endpoint=os.environ['VPF730_ENDPOINT'],
            api_key=os.environ['VPF730_API_KEY'],
        )

    @classmethod
    def from_file(cls, path: str) -> Config:
        """Constructs a new :func:`Config` from a provided ``.ini`` config
        file with this format:

            .. highlight:: ini
            .. code-block:: ini

                [vpf_730]
                local_db=local.db
                queue_db=queue.db
                serial_port=/dev/ttyS0
                endpoint=http://localhost:5000/vpf-730
                api_key=deadbeef

        :param path: path to the ``.ini`` config file with the structure above

        :return: a new instance of :func:`Config` created from a config file
        """
        config = configparser.ConfigParser()
        config.read(path)
        return cls(**dict(config['vpf_730']))

    @classmethod
    def from_argparse(cls, args: argparse.Namespace) -> Config:
        """Constructs a new :func:`Config` from a :func:`argparse.Namespace`,
        created by the argument parser returned by
        :func:`vpf_730.main.build_parser`.

        :param args: arguments returned from the argument parser created by
            :func:`vpf_730.main.build_parser`

        :return: a new instance of :func:`Config` created from CLI arguments
        """
        return cls(
            local_db=args.local_db,
            queue_db=args.queue_db,
            serial_port=args.serial_port,
            endpoint=args.endpoint,
            api_key=os.environ['VPF730_API_KEY'],
        )

    def __repr__(self) -> str:
        return (
            f'{type(self).__name__}(local_db={self.local_db!r}, '
            f'queue_db={self.queue_db!r}, serial_port={self.serial_port!r}, '
            f'endpoint={self.endpoint!r}, api_key=***)'
        )


class Worker(threading.Thread):
    """class representing a worker running in a thread. Please also the see
    python documentation for ``threading.Thread``
    (https://docs.python.org/3/library/threading.html#thread-objects).

    :param queue: a :func:`vpf_730.fifo_queue.Queue` instance the worker should
        get messages from
    :param cfg: a :func:`Config` object providing all information
    :param group: should be None; reserved for future extension when a
        ``ThreadGroup`` class is implemented.
    :param target: **unused**
    :param name: the thread name
    :param args: **unused**
    :param kwargs: **unused**
    :param poll_interval: interval in seconds the worker should poll for
        messages  (default: ``.1``)
    :param daemon: thread is started as a ``daemon``
    """

    def __init__(
            self,
            queue: Queue,
            cfg: Config,
            group: None = None,
            target: Callable[..., Any] | None = None,
            name: str | None = None,
            args: Iterable[Any] = (),
            kwargs: Mapping[str, Any] | None = None,
            *,
            poll_interval: float = .1,
            daemon: bool | None = None,
    ) -> None:
        super().__init__(group, target, name, args, kwargs, daemon=daemon)
        self.running = True
        self.queue = queue
        self.poll_interval = poll_interval
        self.cfg = cfg

    def run(self) -> None:
        """Worker infinity loop, polling and processing messages.

        Setting the attribute ``Worker.running = False`` will stop the worker
        after finishing the currently running task. The polling interval of the
        worker can be set using ``Worker.poll_interval``.
        """
        try:
            while self.running is True:
                if self.queue.empty():
                    time.sleep(self.poll_interval)
                else:
                    msg = self.queue.get()
                    # if the queue is not empty, msg can't be None
                    assert msg is not None
                    try:
                        call = TASKS[msg.task]
                        call(msg, self.cfg)
                        self.queue.task_done(msg)
                    except Exception as e:
                        print(' worker encountered an Error '. center(79, '='))
                        print(f'==> tried processing: {msg}')
                        print(
                            f'====> Traceback:\n'
                            f"{textwrap.indent(traceback.format_exc(), '  ')}",
                        )
                        if SENTRY:  # pragma: no cover
                            sentry_sdk.capture_exception(e)
                        self.queue.task_failed(msg)
        finally:
            del self._target  # type: ignore [attr-defined]
            del self._args, self._kwargs  # type: ignore [attr-defined]

    def finish_and_join(self) -> None:
        """Finish all tasks that are still waiting in the queue and then join
        the thread. This does not apply to tasks having an `eta` in the future.
        Only currently visible messages are considered.
        """
        while not self.queue.empty():
            time.sleep(self.poll_interval)

        self.running = False
        self.join()



from flask import Flask
from flask import request
from flask import redirect
from flask import url_for

app = Flask(__name__)

@app.route('/api/vpf-730', methods=['GET', 'POST'])
def route():
    print(request.json)
    return {'code': 200}

@app.route('/<path:page>', methods=['GET', 'POST'])
def other(page):
    return redirect(url_for('route'))

if __name__ == '__main__':
    app.run(debug=True)



from __future__ import annotations

import contextlib
import json
import sqlite3
from collections.abc import Generator
from datetime import datetime
from datetime import timezone
from typing import Literal
from typing import NamedTuple
from uuid import UUID

from vpf_730.vpf_730 import Measurement


@contextlib.contextmanager
def connect(db_path: str) -> Generator[sqlite3.Connection, None, None]:
    """Context manager to connect to a sqlite database.

    :param db_path: path to the sqlite database

    :return: A Generator yielding an open sqlite connection
    """
    with contextlib.closing(
        sqlite3.connect(
            db_path,
            detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES,
        ),
    ) as db:
        with db:
            yield db


class Message(NamedTuple):
    """A ``NamedTuple`` class representing a Message which can be enqueued.

    :param id: unique id of the message, should be :func:`uuid.uuid4`
    :param task: the name of a registered function i.e. ``my_func.__name__``.
        The function can be registered using
        the :func:`vpf_730.worker.register` decorator.
    :param blob: ``NamedTuple`` (:func:`vpf_730.vpf_730.Measurement`)
        representing a measurement from the VPF-730 sensor.
    :param retries: number of times the messages was unsuccessfully picked up
        by a worker. This is automatically managed by the queue. Calling
        :func:`Queue.task_failed` will increase this in the database and the
        next time the message is fetched it will be updated.
    :param eta: ETA (estimated time of arrival), a unix-timestamp in
        milliseconds which lets you set a specific date and time that is the
        earliest time at which the task will become visible to workers and can
        be executed. If a message should become visible instantly, set it to
        ``None``.
    """
    id: UUID
    # TODO: this should become a callable, serializing: callable.__name__
    task: str
    blob: Measurement
    retries: int = 0
    eta: int | None = None

    def serialize(self) -> dict[str, str | int | None]:
        """serialize the the :func:`Message` ``NamedTuple`` to a dictionary.
            The blob containing :func:`vpf_730.vpf_730.Measurement` is
            serialized to a string representation of a ``json``. This is used
            internally when calling :func:`Queue.put`.

        :return: A serialized version of the :func:`Message` as a dictionary,
            for insertion into a sqlite db
        """
        return {
            'id': self.id.hex,
            'task': self.task,
            'blob': json.dumps(self.blob._asdict()),
            'retries': self.retries,
            'eta': self.eta,
        }

    @classmethod
    def from_queue(cls, msg: tuple[str, str, str, int, int | None]) -> Message:
        """Constructs a new :func:`Message` from the result of a db query to
        the queue. It is used internally when executing :func:`Queue.get`.

        :param msg: a tuple which is the result of a query to the db and
            representing ``[id, task blob, retries, eta]``

        :return: a new instance of :func:`Message`.
        """
        return cls(
            id=UUID(msg[0]),
            task=msg[1],
            blob=Measurement(**json.loads(msg[2])),
            retries=msg[3],
            eta=msg[4],
        )


class Queue():
    r"""A class representing a FIFO message queue

    :param db: a string with the path to the sqlite database
    :param max_retries: the maximum number of times a messages should be
        retried before it is put into a deadletter queue (default: 5)
    :param exponential_backoff: exponent to increase the time between retries.
        This is used to set the ``eta`` when returning the message back to the
        queue after a failure. It is calculated this way:
        :math:`t = n^{exponential\_backoff}`.
    :param keep_msg: the number of (successfully processed) messages to keep in
        the queue database before pruning them (default: 10000)
    :param prune_interval: after how many messages that were :func:`Queue.put`
        into the queue, should the queue-database be pruned (default: 1000)
    """

    def __init__(
            self,
            db: str,
            *,
            max_retries: int = 5,
            exponential_backoff: int | None = None,
            keep_msg: int = 10000,
            prune_interval: int = 1000,
    ) -> None:
        self.db = db
        self.max_retries = max_retries
        self.exponential_backoff = exponential_backoff
        self.keep_msg = keep_msg
        self.prune_interval = prune_interval
        self._nr_puts = 0

        create_queue = '''\
            CREATE TABLE IF NOT EXISTS queue(
                id VARCHAR(36) PRIMARY KEY,
                task TEXT,
                enqueued INT NOT NULL,
                fetched INT,
                acked INT,
                blob JSON NOT NULL,
                retries INT DEFAULT 0 NOT NULL,
                eta INT
            )
        '''
        create_queue_idx = '''\
            CREATE INDEX IF NOT EXISTS idx_queue_enqueued
            ON queue(enqueued)
        '''
        create_deadletter = '''\
            CREATE TABLE IF NOT EXISTS deadletter(
                id VARCHAR(36) PRIMARY KEY,
                task TEXT,
                enqueued INT NOT NULL,
                fetched INT,
                acked INT,
                blob JSON NOT NULL,
                retries INT DEFAULT 0 NOT NULL,
                eta INT
            )
        '''
        with connect(self.db) as con:
            con.execute(create_queue)
            con.execute(create_deadletter)
            con.execute(create_queue_idx)

    def put(
            self,
            msg: Message,
            route: Literal['queue', 'deadletter'] = 'queue',
    ) -> UUID:
        """Add a new message to the queue (determined by ``route``).

        :param msg: a :func:`Message`, that should be added to the queue.
        :param route: in which queue should the message be put. Allowed options
            are ``'queue'`` or ``'deadletter'`` (default: ``queue``)

        :return: the uuid of the message created
        """
        if route == 'queue':
            queue_params = {
                'enqueued': int(datetime.now(timezone.utc).timestamp() * 1000),
            }
            insert = '''\
                INSERT INTO queue(id, task, enqueued, blob, retries, eta)
                VALUES(:id, :task, :enqueued, :blob, :retries, :eta)
            '''
        elif route == 'deadletter':
            # TODO: we should keep the initial enqueued
            queue_params = {
                'enqueued': int(datetime.now(timezone.utc).timestamp() * 1000),
            }
            insert = '''\
                INSERT INTO deadletter(id, task, enqueued, blob, retries, eta)
                VALUES(:id, :task, :enqueued, :blob, :retries, :eta)
            '''
        else:
            raise NotImplementedError

        query_params = msg.serialize() | queue_params
        with connect(self.db) as db:
            db.execute(insert, query_params)

        self._nr_puts += 1
        return msg.id

    def get(
            self,
            route: Literal['queue', 'deadletter'] = 'queue',
    ) -> Message | None:
        """Get a message from a queue (determined by ``route``). If no message
        is available, ``None`` is returned instead of a :func:`Message`.
        Availability of a :func:`Message` is also determined by ``eta``. If the
        ``eta`` of a :func:`Message` lies in the future, it cannot be fetched
        using :func:`Queue.get`.

        :param route: from which queue should the message be polled. Allowed
            options are ``'queue'`` or ``'deadletter'`` (default: ``queue``)

        :return: a :func:`Message` or ``None``
        """
        if route == 'queue':
            get = '''\
                SELECT id, task, blob, retries, eta FROM queue
                WHERE fetched IS NULL AND (eta IS NULL OR eta <= ?)
                ORDER BY enqueued LIMIT 1
            '''
            ts_now_ms = int(datetime.now(timezone.utc).timestamp() * 1000)
            with connect(self.db) as db:
                ret = db.execute(get, (ts_now_ms,))
                val = ret.fetchone()

        elif route == 'deadletter':
            # we ignore eta in deadletter on purpose, it'll become active again
            # when requeued into queue
            get = '''\
                SELECT id, task, blob, retries, eta FROM deadletter
                ORDER BY enqueued LIMIT 1
            '''
            with connect(self.db) as db:
                ret = db.execute(get)
                val = ret.fetchone()

        else:
            raise NotImplementedError

        if val is None:
            return val

        msg = Message.from_queue(val)
        with connect(self.db) as db:
            db.execute(
                'UPDATE queue SET fetched = ? WHERE id = ?',
                (
                    int(datetime.now(timezone.utc).timestamp() * 1000),
                    msg.id.hex,
                ),
            )
        return msg

    def task_done(self, msg: Message) -> None:
        """Mark a task (:func:`Message`) as done.

        :param msg: the :func:`Message` to be marked as done
        """
        with connect(self.db) as db:
            db.execute(
                'UPDATE queue SET acked = ? WHERE id = ?',
                (
                    int(datetime.now(timezone.utc).timestamp() * 1000),
                    msg.id.hex,
                ),
            )

        if self._nr_puts >= self.prune_interval:
            self._prune()

    def task_failed(self, msg: Message) -> None:
        """Mark a task (:func:`Message`) as failed. If the number of
        ``retires`` exceed the maximum number of retires
        (``Queue.max_retires``) the :func:`Message` is routed to the
        ``deadletter`` queue. Otherwise it is returned back to the ``queue``.

        :param msg: the :func:`Message` to be marked as failed
        """
        if msg.retries >= self.max_retries:
            # route to deadletter
            with connect(self.db) as db:
                db.execute('DELETE FROM queue WHERE id = ?', (msg.id.hex,))

            self.put(msg=msg, route='deadletter')
        else:
            # increase number of retries, return back to queue
            eta = None
            if self.exponential_backoff is not None:
                eta = int(datetime.now(timezone.utc).timestamp() * 1000)
                add_ms = ((msg.retries + 1) ** self.exponential_backoff) * 1000
                eta += add_ms

            with connect(self.db) as db:
                db.execute(
                    '''\
                    UPDATE queue
                    SET retries = ?, fetched = NULL, eta = ?
                    WHERE id = ?
                    ''',
                    (msg.retries + 1, eta, msg.id.hex),
                )

    def qsize(self) -> int:
        """Get the current queue size, meaning the number of :func:`Message`
        that are available to consumers in ``queue``.

        :return: number of messages that are ready to be picked up
        """
        with connect(self.db) as db:
            ret = db.execute(
                '''\
                SELECT count(1) FROM queue
                WHERE fetched IS NULL AND (eta IS NULL OR eta <= ?)
                ''',
                (int(datetime.now(timezone.utc).timestamp() * 1000),),
            )
            return ret.fetchone()[0]

    def deadletter_qsize(self) -> int:
        """Get the current deadletter queue size, meaning the total number of
        :func:`Message` in ``deadletter``.

        :return: number of :func:`Message` in deadletter
        """
        with connect(self.db) as db:
            ret = db.execute(
                '''\
                SELECT count(1) FROM deadletter
                WHERE fetched IS NULL AND (eta IS NULL OR eta <= ?)
                ''',
                (int(datetime.now(timezone.utc).timestamp() * 1000),),
            )
            return ret.fetchone()[0]

    def empty(self) -> bool:
        """Boolean indicating if there are messages in ``queue``.

        :return: ``True`` if ``queue`` is empty otherwise ``False``
        """
        return self.qsize() == 0

    def deadletter_empty(self) -> bool:
        """A boolean indicating if there are messages in ``deadletter``.

        :return: ``True`` if ``deadletter`` is empty otherwise ``False``
        """

        return self.deadletter_qsize() == 0

    def deadletter_requeue(self) -> None:
        """Requeue all messages that are in ``deadletter``. Each message is put
        back into ``queue`` and removed from ``deadletter``. Retries are reset
        to 0, meaning each message is tried ``Queue.max_retries`` again.
        """
        while not self.deadletter_empty():
            msg = self.get(route='deadletter')
            # msg can't be None if there are still messages in queue
            assert msg is not None
            # reset the number of retries
            msg = msg._replace(retries=0)
            self.put(msg=msg)
            # remove from deadletter
            with connect(self.db) as db:
                db.execute(
                    'DELETE FROM deadletter WHERE id = ?',
                    (msg.id.hex,),
                )

    def _prune(self) -> None:
        """Prune the ``queue`` table to the latest ``Queue.keep_msg`` messages
        and execute a ``VACUUM`` of the database table.
        """
        with connect(self.db) as db:
            db.execute(
                '''\
                DELETE FROM queue WHERE id NOT IN (
                    SELECT id FROM queue
                    WHERE acked IS NOT NULL
                    ORDER BY enqueued DESC
                    LIMIT ?
                ) AND acked IS NOT NULL
                ''',
                (self.keep_msg,),
            )
        # VACUUM needs a separate transaction
        with connect(self.db) as db:
            db.execute('VACUUM')

        self._nr_puts = 0
